#
# Copyright (c) 2002, 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# arc class
#

from __future__ import generators

import math

import style
import util
import graphicobject
import entity
import linetype
import color
import point
import tolerance
import quadtree

_dtr = math.pi/180.0
_rtd = 180.0/math.pi

class Arc(graphicobject.GraphicObject):
    """A class for Arcs.

An Arc has four attributes:

center: A Point object
radius: The Arc's radius
start_angle: The start angle
end_angle: The end angle

An Arc has the following methods:

{get/set}Center(): Get/Set the center Point of an Arc.
{get/set}Radius(): Get/Set the radius of an Arc.
{get/set}StartAngle(): Get/Set the starting angle of an Arc.
{get/set}EndAngle(): Get/Set the end angle of an arc
move(): Move the Arc.
length(): Return the length of an Arc.
area(): Return the area of an Arc.
mapPoint(): Find the nearest Point on the Arc to some other Point.
mapCoords(): Find the nearest Point on the Arc to a coordinate pair.
inRegion(): Returns whether or not a Arc can be seen in a bounded area.
clone(): Return an indentical copy of a Arc.
    """

    __defstyle = style.Style(u'Default Arc Style',
                            linetype.Linetype(u'Solid', None),
                            color.Color(0xffffff),
                            1.0)

    messages = {
        'moved' : True,
        'center_changed' : True,
        'radius_changed' : True,
        'start_angle_changed' : True,
        'end_angle_changed' : True
        }

    def __init__(self, center, radius, start_angle, end_angle,
                 st=None, lt=None, col=None, th=None, **kw):
        """Initialize a Arc.

Arc(center, radius, start_angle, end_angle)

The center should be a Point, or a two-entry tuple of floats,
and the radius should be a float greater than 0.
        """
        _cp = center
        if not isinstance(_cp, point.Point):
            _cp = point.Point(center)
        _r = radius
        if not isinstance(_r, float):
            _r = float(radius)
        if not _r > 0.0:
            raise ValueError, "Invalid radius: %g" % _r
        _st = st
        _sa = util.make_c_angle(start_angle)
        _ea = util.make_c_angle(end_angle)
        if _st is None:
            _st = Arc.__defstyle
        graphicobject.GraphicObject.__init__(self, _st, lt, col, th, **kw)
        self.__radius = _r
        self.__sa = _sa
        self.__ea = _ea
        self.__center = _cp
        _cp.connect('moved', self, Arc.movePoint)
        _cp.storeUser(self)

    def __eq__(self, obj):
        """Compare a Arc to another for equality.
        """
        if not isinstance(obj, Arc):
            return False
        _val = False
        if obj is self:
            _val = True
        else:
            if ((self.__center == obj.getCenter()) and
                (abs(self.__radius - obj.getRadius()) < 1e-10) and
                (abs(self.__sa - obj.getStartAngle()) < 1e-10) and
                (abs(self.__ea - obj.getEndAngle()) < 1e-10)):
                    _val = True
        return _val

    def __ne__(self, obj):
        """Compare a Arc to another for inequality.
        """
        if not isinstance(obj, Arc):
            return True
        _val = True
        if obj is self:
            _val = False
        else:
            if ((self.__center == obj.getCenter()) and
                (abs(self.__radius - obj.getRadius()) < 1e-10) and
                (abs(self.__sa - obj.getStartAngle()) < 1e-10) and
                (abs(self.__ea - obj.getEndAngle()) < 1e-10)):
                    _val = False
        return _val

    def getValues(self):
        _pids = entity.Entity.getValues(self)
        _gv = self.getGraphicValues(Arc.__defstyle)
        return ('arc', _pids, _gv, self.__center.getID(), self.__radius,
                self.__sa, self.__ea)

    def getCenter(self):
        """Return the center Point of the Arc.

getCenter()
        """
        return self.__center

    def setCenter(self, c):
        """Set the center Point of the Arc.

setCenter(c)

The argument must be a Point or a tuple containing
two float values.
        """
        if self.isLocked():
            raise RuntimeError, "Setting center not allowed - object locked."
        _cp = self.__center
        if not isinstance(c, point.Point):
            raise TypeError, "Invalid center point: " + `c`
        if _cp is not c:
            _cp.disconnect(self)
            _cp.freeUser(self)
            self.__center = c
            self.sendMessage('center_changed', _cp)
            c.connect('moved', self, Arc.movePoint)
            c.storeUser(self)
            if abs(_cp.x - c.x) > 1e-10 or abs(_cp.y - c.y) > 1e-10:
                self.sendMessage('moved', _cp.x, _cp.y, self.__radius,
                                 self.__sa, self.__ea)
            self.modified()

    center = property(getCenter, setCenter, None, "Arc center")

    def getRadius(self):
        """Return the radius of the the Arc.

getRadius()
        """
        return self.__radius

    def setRadius(self, radius):
        """Set the radius of the Arc.

setRadius(radius)

The argument must be float value greater than 0.
        """
        if self.isLocked():
            raise RuntimeError, "Setting radius not allowed - object locked."
        _r = radius
        if not isinstance(_r, float):
            _r = float(radius)
        if not _r > 0.0:
            raise ValueError, "Invalid radius: %g" % _r
        _cr = self.__radius
        if abs(_cr - _r) > 1e-10:
            self.__radius = _r
            self.sendMessage('radius_changed', _cr)
            _cx, _cy = self.__center.getCoords()
            self.sendMessage('moved', _cx, _cy, _cr, self.__sa, self.__ea)
            self.modified()

    radius = property(getRadius, setRadius, None, "Arc radius")

    def getStartAngle(self):
        """Return the start_angle for the Arc.

getStartAngle()
        """
        return self.__sa

    def setStartAngle(self, angle):
        """Set the start_angle for the Arc.

setStartAngle(angle)

The argument angle should be a float.
        """
        if self.isLocked():
            raise RuntimeError, "Setting start angle not allowed - object locked."
        _sa = self.__sa
        _angle = util.make_c_angle(angle)
        if abs(_sa - _angle) > 1e-10:
            self.__sa = _angle
            self.sendMessage('start_angle_changed', _sa)
            _cx, _cy = self.__center.getCoords()
            self.sendMessage('moved', _cx, _cy, self.__radius, _sa, self.__ea)
            self.modified()

    start_angle = property(getStartAngle, setStartAngle, None,
                           "Start angle for the Arc.")

    def getEndAngle(self):
        """Return the end_angle for the Arc.

getEndAngle()
        """
        return self.__ea

    def setEndAngle(self, angle):
        """Set the end_angle for the Arc.

setEndAngle(angle)

The argument angle should be a float.
        """
        if self.isLocked():
            raise RuntimeError, "Setting end angle not allowed - object locked."
        _ea = self.__ea
        _angle = util.make_c_angle(angle)
        if abs(_ea - _angle) > 1e-10:
            self.__ea = _angle
            self.sendMessage('end_angle_changed', _ea)
            _cx, _cy = self.__center.getCoords()
            self.sendMessage('moved', _cx, _cy, self.__radius, self.__sa, _ea)
            self.modified()

    end_angle = property(getEndAngle, setEndAngle, None,
                         "End angle for the Arc.")

    def getAngle(self):
        """Return the angular sweep of the Arc.

getAngle()
        """
        _sa = self.__sa
        _ea = self.__ea
        if abs(_ea - _sa) < 1e-10:
            _angle = 360.0
        elif _ea > _sa:
            _angle = _ea - _sa
        else:
            _angle = 360.0 - _sa + _ea
        return _angle

    def throughAngle(self, angle):
        """Return True if an arc passes through some angle

throughAngle(angle)

The argument angle should be a float value. This method returns
True if the arc exists at that angle, otherwise the method returns False.
        """
        _angle = angle
        if not isinstance(_angle, float):
            _angle = float(angle)
        _angle = math.fmod(_angle, 360.0)
        if _angle < 0.0:
            _angle = _angle + 360.0
        _sa = self.__sa
        _ea = self.__ea
        _val = True
        if abs(_sa - _ea) > 1e-10:
            if _sa > _ea:
                if _angle > _ea and _angle < _sa:
                    _val = False
            else:
                if _angle > _ea or _angle < _sa:
                    _val = False
        return _val

    def getEndpoints(self):
        """Return where the two endpoints for the arc-segment lie.

getEndpoints(self)

This function returns two tuples, each containing the x-y coordinates
of the arc endpoints. The first tuple corresponds to the endpoint at
the start_angle, the second to the endpoint at the end_angle.
        """
        _cx, _cy = self.__center.getCoords()
        _r = self.__radius
        _sa = self.__sa
        _sax = _cx + _r * math.cos(_sa * _dtr)
        _say = _cy + _r * math.sin(_sa * _dtr)
        _ea = self.__ea
        _eax = _cx + _r * math.cos(_ea * _dtr)
        _eay = _cy + _r * math.sin(_ea * _dtr)
        return (_sax, _say), (_eax, _eay)

    def length(self):
        """Return the length of the Arc.

length()
        """
        return 2.0 * math.pi * self.__radius * (self.getAngle()/360.0)

    def area(self):
        """Return the area enclosed by the Arc.

area()
        """
        return math.pi * pow(self.__radius, 2) * (self.getAngle()/360.0)

    def move(self, dx, dy):
        """Move a Arc.

move(dx, dy)

The first argument gives the x-coordinate displacement,
and the second gives the y-coordinate displacement. Both
values should be floats.
        """
        if self.isLocked():
            raise RuntimeError, "Setting radius not allowed - object locked."
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        if abs(_dx) > 1e-10 or abs(_dy) > 1e-10:
            _x, _y = self.__center.getCoords()
            self.ignore('moved')
            self.__center.move(_dx, _dy)
            self.receive('moved')
            self.sendMessage('moved', _x, _y, self.__radius,
                             self.__sa, self.__ea)
            self.modified()


    def mapCoords(self, x, y, tol=tolerance.TOL):
        """Return the nearest Point on the Arc to a coordinate pair.

mapCoords(x, y[, tol])

The function has two required arguments:

x: A Float value giving the x-coordinate
y: A Float value giving the y-coordinate

There is a single optional argument:

tol: A float value equal or greater than 0.0

This function is used to map a possibly near-by coordinate pair to
an actual Point on the Arc. If the distance between the actual
Point and the coordinates used as an argument is less than the tolerance,
the actual Point is returned. Otherwise, this function returns None.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _cx, _cy = self.__center.getCoords()
        _r = self.__radius
        _dist = math.hypot((_x - _cx), (_y - _cy))
        if abs(_dist - _r) < _t:
            _angle = _rtd * math.atan2((_y - _cy),(_x - _cx))
            if _angle < 0.0:
                _angle = _angle + 360.0
            if self.throughAngle(_angle):
                _xoff = _r * math.cos(_angle)
                _yoff = _r * math.sin(_angle)
                return (_cx + _xoff), (_cy + _yoff)
        return None

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not an Arc exists within a region.

inRegion(xmin, ymin, xmax, ymax[, fully])

The first four arguments define the boundary. The optional
fifth argument fully indicates whether or not the Arc
must be completely contained within the region or just pass
through it.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + `fully`
        _xc, _yc = self.__center.getCoords()
        _r = self.__radius
        #
        # cheap test to see if arc cannot be in region
        #
        _axmin, _aymin, _axmax, _aymax = self.getBounds()
        if ((_axmin > _xmax) or
            (_aymin > _ymax) or
            (_axmax < _xmin) or
            (_aymax < _ymin)):
            return False
        _val = False
        _bits = 0
        #
        # calculate distances from center to region boundary
        #
        if abs(_xc - _xmin) < _r: _bits = _bits | 1 # left edge
        if abs(_xc - _xmax) < _r: _bits = _bits | 2 # right edge
        if abs(_yc - _ymin) < _r: _bits = _bits | 4 # bottom edge
        if abs(_yc - _ymax) < _r: _bits = _bits | 8 # top edge
        if _bits == 0:
            #
            # circle must be visible - the center is in
            # the region and is more than the radius from
            # each edge
            #
            _val = True
        else:
            #
            # calculate distance to corners of region
            #
            if math.hypot((_xc - _xmin), (_yc - _ymax)) < _r:
                _bits = _bits | 0x10 # upper left
            if math.hypot((_xc - _xmax), (_yc - _ymin)) < _r:
                _bits = _bits | 0x20 # lower right
            if math.hypot((_xc - _xmin), (_yc - _ymin)) < _r:
                _bits = _bits | 0x40 # lower left
            if math.hypot((_xc - _xmax), (_yc - _ymax)) < _r:
                _bits = _bits | 0x80 # upper right
            #
            # if all bits are set then distance from circle center
            # to region endpoints is less than radius - circle
            # entirely outside the region
            #
            _val = not ((_bits == 0xff) or fully)
            #
            # if the test value is still true, check that the
            # arc boundary can overlap with the region
            #
            if _val:
                _sa = self.__sa
                _ea = self.__ea
                _xs = _xc + (_r * math.cos(_sa * _dtr))
                _ys = _yc + (_r * math.sin(_sa * _dtr))
                _xe = _xc + (_r * math.cos(_ea * _dtr))
                _ye = _xc + (_r * math.sin(_ea * _dtr))
                _axmin = min(_xc, _xs, _xe)
                if self.throughAngle(180.0):
                    _axmin = _xc - _r
                if _axmin > _xmax:
                    return False
                _aymin = min(_yc, _ys, _ye)
                if self.throughAngle(270.0):
                    _aymin = _yc - _r
                if _aymin > _ymax:
                    return False
                _axmax = max(_xc, _xs, _xe)
                if self.throughAngle(0.0):
                    _axmax = _xc + _r
                if _axmax < _xmin:
                    return False
                _aymax = max(_yc, _ys, _ye)
                if self.throughAngle(90.0):
                    _aymax = _yc + _r
                if _aymax < _ymin:
                    return False
        return _val

    def getBounds(self):
        _xc, _yc = self.__center.getCoords()
        _r = self.__radius
        _sa = self.__sa
        _ea = self.__ea
        _xs = _xc + (_r * math.cos(_sa * _dtr))
        _ys = _yc + (_r * math.sin(_sa * _dtr))
        _xe = _xc + (_r * math.cos(_ea * _dtr))
        _ye = _yc + (_r * math.sin(_ea * _dtr))
        _xmin = min(_xc, _xs, _xe)
        _ymin = min(_yc, _ys, _ye)
        _xmax = max(_xc, _xs, _xe)
        _ymax = max(_yc, _ys, _ye)
        if self.throughAngle(0.0):
            _xmax = _xc + _r
        if self.throughAngle(90.0):
            _ymax = _yc + _r
        if self.throughAngle(180.0):
            _xmin = _xc - _r
        if self.throughAngle(270.0):
            _ymin = _yc - _r
        return _xmin, _ymin, _xmax, _ymax

    def movePoint(self, p, *args):
        _alen = len(args)
        if _alen < 2:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        _cp = self.__center
        if p is not _cp:
            raise ValueError, "Point is not arc center: " + `p`
        _x, _y = _cp.getCoords()
        self.sendMessage('moved', _x, _y, self.__radius, self.__sa, self.__ea)
        self.modified()

    def clone(self):
        """Create an identical copy of a Arc

clone()
        """
        _cp = self.__center.clone()
        _r = self.__radius
        _sa = self.__sa
        _ea = self.__ea
        _st = self.getStyle()
        _lt = self.getLinetype()
        _col = self.getColor()
        _th = self.getThickness()
        return Arc(_cp, _r, _sa, _ea, _st, _lt, _col, _th)

    def sendsMessage(self, m):
        if m in Arc.messages:
            return True
        return graphicobject.GraphicObject.sendsMessage(self, m)


#
# static functions for Arc class
#

    def test_angle(s, e, a):
        """Returns if an angle lies between the start and end angle of an arc.

test_angle(s, e, a)

The arguments are:

s: arc start angle
e: arc end angle
a: angle being tested
        """
        _val = False
        if ((abs(e - s) < 1e-10) or
            ((s > e) and
             ((s <= a <= 360.0) or (0.0 <= a <= e))) or
            (s <= a <= e)):
            _val = True
        return _val

    test_angle = staticmethod(test_angle)

#
# Quadtree Arc storage
#

class ArcQuadtree(quadtree.Quadtree):
    def __init__(self):
        quadtree.Quadtree.__init__(self)

    def getNodes(self, *args):
        _alen = len(args)
        if _alen != 4:
            raise ValueError, "Expected 4 arguments, got %d" % _alen
        _axmin, _aymin, _axmax, _aymax = args
        if not isinstance(_axmin, float):
            _axmin = float(args[0])
        if not isinstance(_aymin, float):
            _aymin = float(args[1])
        if not isinstance(_axmax, float):
            _axmax = float(args[2])
        if not isinstance(_aymax, float):
            _aymax = float(args[3])
        if not _axmax > _axmin:
            raise ValueError, "xmax not greater than xmin"
        if not _aymax > _aymin:
            raise ValueError, "ymax not greater than ymin"
        _nodes = [self.getTreeRoot()]
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_axmin > _xmax) or
                (_axmax < _xmin) or
                (_aymin > _ymax) or
                (_aymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _axmax < _xmid: # arc on left side
                    _ne = _se = False
                if _axmin > _xmid: # arc on right side
                    _nw = _sw = False
                if _aymax < _ymid: # arc below
                    _nw = _ne = False
                if _aymin > _ymid: # arc above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                yield _node

    def addObject(self, obj):
        if not isinstance(obj, Arc):
            raise TypeError, "Invalid Arc object: " + `obj`
        if obj in self:
            return
        _bounds = self.getTreeRoot().getBoundary()
        _xmin = _ymin = _xmax = _ymax = None
        _axmin, _aymin, _axmax, _aymax = obj.getBounds()
        _resize = False
        if _bounds is None: # first node in tree
            _resize = True
            _xmin = _axmin - 1.0
            _ymin = _aymin - 1.0
            _xmax = _axmax + 1.0
            _ymax = _aymax + 1.0
        else:
            _xmin, _ymin, _xmax, _ymax = _bounds
            if _axmin < _xmin:
                _xmin = _axmin - 1.0
                _resize = True
            if _axmax > _xmax:
                _xmax = _axmax + 1.0
                _resize = True
            if _aymin < _ymin:
                _ymin = _aymin - 1.0
                _resize = True
            if _aymax > _ymax:
                _ymax = _aymax + 1.0
                _resize = True
        if _resize:
            self.resize(_xmin, _ymin, _xmax, _ymax)
        for _node in self.getNodes(_axmin, _aymin, _axmax, _aymax):
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _node.addObject(obj)
        quadtree.Quadtree.addObject(self, obj)
        obj.connect('moved', self, ArcQuadtree.moveArc)

    def old_addObject(self, obj):
        if not isinstance(obj, Arc):
            raise TypeError, "Invalid Arc object: " + `obj`
        if obj in self:
            return
        _x, _y = obj.getCenter().getCoords()
        _node = self.getTreeRoot()
        _bounds = _node.getBoundary()
        _xmin = _ymin = _xmax = _ymax = None
        _axmin, _aymin, _axmax, _aymax = obj.getBounds()
        _resize = False
        if _bounds is None: # first node in tree
            _resize = True
            _xmin = _axmin - 1.0
            _ymin = _aymin - 1.0
            _xmax = _axmax + 1.0
            _ymax = _aymax + 1.0
        else:
            _xmin, _ymin, _xmax, _ymax = _bounds
            if _axmin < _xmin:
                _xmin = _axmin - 1.0
                _resize = True
            if _axmax > _xmax:
                _xmax = _axmax + 1.0
                _resize = True
            if _aymin < _ymin:
                _ymin = _aymin - 1.0
                _resize = True
            if _aymax > _ymax:
                _ymax = _aymax + 1.0
                _resize = True
        if _resize:
            self.resize(_xmin, _ymin, _xmax, _ymax)
            _node = self.getTreeRoot()
        _nodes = []
        _nodes.append(_node)
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_axmin > _xmax) or
                (_axmax < _xmin) or
                (_aymin > _ymax) or
                (_aymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _axmax < _xmid: # arc on left side
                    _ne = _se = False
                if _axmin > _xmid: # arc on right side
                    _nw = _sw = False
                if _aymax < _ymid: # arc below
                    _nw = _ne = False
                if _aymin > _ymid: # arc above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                if obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                    _node.addObject(obj)
        quadtree.Quadtree.addObject(self, obj)
        obj.connect('moved', self, ArcQuadtree.moveArc)

    def delObject(self, obj):
        if obj not in self:
            return
        _axmin, _aymin, _axmax, _aymax = obj.getBounds()
        _pdict = {}
        for _node in self.getNodes(_axmin, _aymin, _axmax, _aymax):
            _node.delObject(obj) # arc may not be in the node ...
            _parent = _node.getParent()
            if _parent is not None:
                _pid = id(_parent)
                if _pid not in _pdict:
                    _pdict[_pid] = _parent
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        for _parent in _pdict.values():
            self.purgeSubnodes(_parent)

    def old_delObject(self, obj):
        if obj not in self:
            return
        _x, _y = obj.getCenter().getCoords()
        _xmin = _ymin = _xmax = _ymax = None
        _axmin, _aymin, _axmax, _aymax = obj.getBounds()
        _pdict = {}
        _nodes = []
        _nodes.append(self.getTreeRoot())
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_axmin > _xmax) or
                (_axmax < _xmin) or
                (_aymin > _ymax) or
                (_aymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _axmax < _xmid: # arc on left side
                    _ne = _se = False
                if _axmin > _xmid: # arc on right side
                    _nw = _sw = False
                if _aymax < _ymid: # arc below
                    _nw = _ne = False
                if _aymin > _ymid: # arc above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                _node.delObject(obj) # arc may not be in the node ...
                _parent = _node.getParent()
                if _parent is not None:
                    _pid = id(_parent)
                    if _pid not in _pdict:
                        _pdict[_pid] = _parent
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        for _parent in _pdict.values():
            self.purgeSubnodes(_parent)

    def find(self, *args):
        _alen = len(args)
        if _alen < 5:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        _r = args[2]
        if not isinstance(_r, float):
            _r = float(args[2])
        _sa = args[3]
        if not isinstance(_sa, float):
            _sa = float(args[3])
        _ea = args[4]
        if not isinstance(_ea, float):
            _ea = float(args[4])
        _t = tolerance.TOL
        if _alen > 5:
            _t = tolerance.toltest(args[5])
        if not len(self):
            return None
        _axmin = _x - _r
        _axmax = _x + _r
        _aymin = _y - _r
        _aymax = _y + _r
        _nodes = []
        _nodes.append(self.getTreeRoot())
        _adict = {}
        _bailout = False
        _arc = None
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_axmin > _xmax) or
                (_axmax < _xmin) or
                (_aymin > _ymax) or
                (_aymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _axmax < (_xmid - _t): # arc on left side
                    _ne = _se = False
                if _axmin > (_xmid + _t): # arc on right side
                    _nw = _sw = False
                if _aymax < (_ymid - _t): # arc below
                    _nw = _ne = False
                if _aymin > (_ymid + _t): # arc above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                for _a in _node.getObjects():
                    _aid = id(_a)
                    if _aid not in _adict:
                        _ax, _ay = _a.getCenter().getCoords()
                        if ((abs(_ax - _x) < _t) and (abs(_ay - _y) < _t)):
                            _ar = _a.getRadius()
                            if abs(_ar - _r) < _t:
                                _asa = _a.getStartAngle()
                                _aea = _a.getEndAngle()
                                if abs(_asa - _sa) < 1e-10 and abs(_aea - _ea) < 1e-10:
                                    _bailout = True
                                    _arc = _a
                                break
                        _adict[_aid] = True
            if _bailout:
                break
        return _arc

    def moveArc(self, obj, *args):
        if obj not in self:
            raise ValueError, "Arc not stored in Quadtree: " + `obj`
        _alen = len(args)
        if _alen < 5:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        _r = args[2]
        if not isinstance(_r, float):
            _r = float(args[2])
        _sa = args[3]
        if not isinstance(_sa, float):
            _sa = float(args[3])
        _ea = args[4]
        if not isinstance(_ea, float):
            _ea = float(args[4])
        _sax = _x + _r * math.cos(_sa * _dtr)
        _say = _y + _r * math.sin(_sa * _dtr)
        _eax = _x + _r * math.cos(_ea * _dtr)
        _eay = _y + _r * math.sin(_ea * _dtr)
        _axmin = min(_x, _sax, _eax)
        if ((abs(_sa - 180.0) < 1e-10) or
            (abs(_ea - 180.0) < 1e-10) or
            ((_sa < _ea) and (_sa < 180.0 < _ea)) or
            ((_ea < _sa) and (_sa < 180.0))):
            _axmin = _x - _r
        _axmax = max(_x, _sax, _eax)
        if ((abs(_sa) < 1e-10) or
            (abs(_ea) < 1e-10) or
            ((_sa > _ea) and (_ea > 0.0))):
            _axmax = _x + _r
        _aymin = min(_y, _say, _eay)
        if ((abs(_sa - 270.0) < 1e-10) or
            (abs(_ea - 270.0) < 1e-10) or
            ((_sa < _ea) and (_sa < 270.0 < _ea)) or
            ((_ea < _sa) and (_sa < 270.0))):
            _axmin = _x - _r
        _aymax = max(_y, _say, _eay)
        if ((abs(_sa - 90.0) < 1e-10) or
            (abs(_ea - 90.0) < 1e-10) or
            ((_sa < _ea) and (_sa < 90.0 < _ea)) or
            ((_ea < _sa) and (_sa < 90.0))):
            _axmin = _x + _r
        for _node in self.getNodes(_axmin, _aymin, _axmax, _aymax):
            _node.delObject(obj) # arc may not be in node ...
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        self.addObject(obj)

    def old_moveArc(self, obj, *args):
        if obj not in self:
            raise ValueError, "Arc not stored in Quadtree: " + `obj`
        _alen = len(args)
        if _alen < 5:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        _r = args[2]
        if not isinstance(_r, float):
            _r = float(args[2])
        _sa = args[3]
        if not isinstance(_sa, float):
            _sa = float(args[3])
        _ea = args[4]
        if not isinstance(_ea, float):
            _ea = float(args[4])
        _axmin = _x - _r
        _axmax = _x + _r
        _aymin = _y - _r
        _aymax = _y + _r
        _nodes = []
        _nodes.append(self.getTreeRoot())
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_axmin > _xmax) or
                (_axmax < _xmin) or
                (_aymin > _ymax) or
                (_aymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _axmax < _xmid: # arc on left side
                    _ne = _se = False
                if _axmin > _xmid: # arc on right side
                    _nw = _sw = False
                if _aymax < _ymid: # arc below
                    _nw = _ne = False
                if _aymin > _ymid: # arc above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                _node.delObject(obj) # arc may not be in node ...
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        self.addObject(obj)

    def getClosest(self, x, y, tol=tolerance.TOL):
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _arc = _tsep = None
        _bailout = False
        _adict = {}
        _nodes = []
        _nodes.append(self.getTreeRoot())
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_x < (_xmin - _t)) or
                (_x > (_xmax + _t)) or
                (_y < (_ymin - _t)) or
                (_y > (_ymax + _t))):
                continue
            if _node.hasSubnodes():
                _nodes.extend(_node.getSubnodes())
            else:
                for _a in _node.getObjects():
                    _aid = id(_a)
                    if _aid not in _adict:
                        _ap = _a.mapCoords(_x, _y, _t)
                        if _ap is not None:
                            _ax, _ay = _ap
                            _sep = math.hypot((_ax - _x), (_ay - _y))
                            if _tsep is None:
                                _tsep = _sep
                                _arc = _a
                            else:
                                if _sep < _tsep:
                                    _tsep = _sep
                                    _arc = _a
                            if _sep < 1e-10 and _arc is not None:
                                _bailout = True
                                break
            if _bailout:
                break
        return _arc

    def getInRegion(self, xmin, ymin, xmax, ymax):
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _arcs = []
        if not len(self):
            return _arcs
        _nodes = []
        _nodes.append(self.getTreeRoot())
        _adict = {}
        while len(_nodes):
            _node = _nodes.pop()
            if _node.hasSubnodes():
                for _subnode in _node.getSubnodes():
                    _sxmin, _symin, _sxmax, _symax = _subnode.getBoundary()
                    if ((_sxmin > _xmax) or
                        (_symin > _ymax) or
                        (_sxmax < _xmin) or
                        (_symax < _ymin)):
                        continue
                    _nodes.append(_subnode)
            else:
                for _arc in _node.getObjects():
                    _aid = id(_arc)
                    if _aid not in _adict:
                        if _arc.inRegion(_xmin, _ymin, _xmax, _ymax):
                            _arcs.append(_arc)
                        _adict[_aid] = True
        return _arcs

#
# Arc history class
#

class ArcLog(graphicobject.GraphicObjectLog):
    def __init__(self, a):
        if not isinstance(a, Arc):
            raise TypeError, "Invalid arc: " + `a`
        graphicobject.GraphicObjectLog.__init__(self, a)
        a.connect('center_changed' ,self, ArcLog.centerChange)
        a.connect('radius_changed', self, ArcLog.radiusChange)
        a.connect('start_angle_changed', self, ArcLog.saChange)
        a.connect('end_angle_changed', self, ArcLog.eaChange)

    def radiusChange(self, a, *args):
        _alen = len(args)
        if _alen < 1:
            raise ValueError, "Invalid argument count: %d" % _alen
        _r = args[0]
        if not isinstance(_r, float):
            _r = float(args[0])
        self.saveUndoData('radius_changed', _r)

    def centerChange(self, a, *args):
        _alen = len(args)
        if _alen < 1:
            raise ValueError, "Invalid argument count: %d" % _alen
        _old = args[0]
        if not isinstance(_old, point.Point):
            raise TypeError, "Invalid old center point: " + `_old`
        _oid = _old.getID()
        self.saveUndoData('center_changed', _oid)

    def saChange(self, a, *args):
        _alen = len(args)
        if _alen < 1:
            raise ValueError, "Invalid argument count: %d" % _alen
        _sa = args[0]
        if not isinstance(_sa, float):
            _sa = float(args[0])
        self.saveUndoData('sa_change', _sa)

    def eaChange(self, a, *args):
        _alen = len(args)
        if _alen < 1:
            raise ValueError, "Invalid argument count: %d" % _alen
        _ea = args[0]
        if not isinstance(_ea, float):
            _ea = float(args[0])
        self.saveUndoData('ea_change', _ea)

    def execute(self, undo, *args):
        if undo is not True and undo is not False:
            raise ValueError, "Invalid undo value: " + str(undo)
        _alen = len(args)
        if _alen == 0:
            raise ValueError, "No arguments to execute()"
        _a = self.getObject()
        _cp = _a.getCenter()
        _op = args[0]
        if _op == 'center_changed':
            if _alen < 2:
                raise ValueError, "Invalid argument count: %d" % _alen
            _oid = args[1]
            _cid = _cp.getID()
            if undo:
                self.saveRedoData('center_changed', _cid)
                self.ignore('center_changed')
                try:
                    _parent = _a.getParent()
                    if _parent is None:
                        raise ValueError, "Arc has no parent - cannot undo"
                    _op = _parent.getObject(_oid)
                    if _op is None or not isinstance(_op, point.Point):
                        raise ValueError, "Center point missing: id=%d" % _oid
                    _x1, _y1 = _cp.getCoords()
                    _a.startUndo(True)
                    try:
                        _a.setCenter(_op)
                    finally:
                        _a.endUndo()
                    if (abs(_op.x - _x1) > 1e-10 or
                        abs(_op.y - _y1) > 1e-10):
                        _a.sendMessage('moved', _x1, _y1, _a.getRadius(),
                                       _a.getStartAngle(), _a.getEndAngle())
                finally:
                    self.receive('center_changed')
            else:
                self.saveUndoData('center_changed', _cid)
                _parent = _a.getParent()
                if _parent is None:
                    raise ValueError, "Arc has no parent - cannot undo"
                _op = _parent.getObject(_oid)
                if _op is None or not isinstance(_op, point.Point):
                    raise ValueError, "Center point missing: id=%d" % _oid
                _ep1, _ep2 = _a.getEndpoints()
                _ep = _parent.find('point', _ep1[0], _ep1[1])
                assert _ep is not None, "Arc endpoint not found in layer"
                _ep.freeUser(_a)
                _ep = _parent.find('point', _ep2[0], _ep2[1])
                assert _ep is not None, "Arc endpoint not found in layer"
                _ep.freeUser(_a)
                self.ignore('center_changed')
                try:
                    _a.setCenter(_op)
                finally:
                    self.receive('center_changed')
        elif _op == 'radius_changed':
            if len(args) < 2:
                raise ValueError, "Invalid argument count: %d" % _alen
            _r = args[1]
            if not isinstance(_r, float):
                _r = float(args[1])
            _cr = _a.getRadius()
            if undo:
                self.saveUndoData('radius_changed', _cr)
            else:
                self.saveRedoData('radius_changed', _cr)
            self.ignore('radius_changed')
            try:
                if undo:
                    _a.startUndo(True)
                    try:
                        _a.setRadius(_r)
                    finally:
                        _a.endUndo()
                    if abs(_r - _cr) > 1e-10:
                        _a.sendMessage('moved', _cp.x, _cp.y, _cr,
                                       _a.getStartAngle(), _a.getEndAngle())
                else:
                    _a.setRadius(_r)
            finally:
                self.receive('radius_changed')
        elif _op == 'sa_change':
            if len(args) < 2:
                raise ValueError, "Invalid argument count: %d" % _alen
            _sa = args[1]
            if not isinstance(_sa, float):
                _sa = float(args[1])
            _csa = _a.getStartAngle()
            if undo:
                self.saveUndoData('sa_change', _csa)
            else:
                self.saveRedoData('sa_change', _csa)
            self.ignore('start_angle_changed')
            try:
                if undo:
                    _a.startUndo(True)
                    try:
                        _a.setStartAngle(_sa)
                    finally:
                        _a.endUndo()
                    if abs(_csa - _sa) > 1e-10:
                        _a.sendMessage('moved', _cp.x, _cp.y, _a.getRadius(),
                                       _csa, _a.getEndAngle())
                else:
                    _a.setStartAngle(_sa)
            finally:
                self.receive('start_angle_changed')
        elif _op == 'ea_change':
            if len(args) < 2:
                raise ValueError, "Invalid argument count: %d" % _alen
            _ea = args[1]
            if not isinstance(_ea, float):
                _ea = float(args[1])
            _cea = _a.getEndAngle()
            if undo:
                self.saveUndoData('ea_change', _cea)
            else:
                self.saveRedoData('ea_change', _cea)
            self.ignore('end_angle_changed')
            try:
                if undo:
                    _a.startUndo(True)
                    try:
                        _a.setEndAngle(_ea)
                    finally:
                        _a.endUndo()
                    if abs(_cea - _ea) > 1e-10:
                        _a.sendMessage('moved', _cp.x, _cp.y, _a.getRadius(),
                                       _a.getStartAngle(), _cea)
                else:
                    _a.setEndAngle(_ea)
            finally:
                self.receive('end_angle_changed')
        else:
            raise ValueError, "Unexpected operation: %s" % _op



#
# Copyright (c) 2002, 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# code to handle moving objects
#

import Generic.point
import Generic.circle
import Generic.arc
import Generic.conobject
import Generic.hcline
import Generic.vcline
import Generic.acline
import Generic.segjoint
import Generic.cline
import Generic.ccircle
import Generic.leader
import Generic.polyline
import Generic.text
import Generic.dimension
import Generic.layer

def _adjust_dimensions(layer, point, newpoint):
    _users = point.getUsers()
    for _uref in _users:
        _user = _uref()
        if _user is not None:
            if isinstance(_user, Generic.dimension.Dimension):
                if isinstance(_user, Generic.dimension.LinearDimension):
                    _l1, _l2 = _user.getDimLayers()
                    _p1, _p2 = _user.getDimPoints()
                    if _l1 is layer and _p1 is point:
                        _user.setP1(layer, newpoint)
                    elif _l2 is layer and _p2 is point:
                        _user.setP2(layer, newpoint)
                elif isinstance(_user, Generic.dimension.AngularDimension):
                    _vl, _l1, _l2 = _user.getDimLayers()
                    _vp, _p1, _p2 = _user.getDimPoints()
                    if _vl is layer and _vp is point:
                        _user.setVertexPoint(layer, newpoint)
                    elif _l1 is layer and _p1 is point:
                        _user.setP1(layer, newpoint)
                    elif _l2 is layer and _p2 is point:
                        _user.setP2(layer, newpoint)
                else:
                    raise TypeError, "Unknown dimension: " + `_user`

def _move_point(layer, point, dx, dy):
    _x, _y = point.getCoords()
    _nx = _x + dx
    _ny = _y + dy
    _np = layer.find('point', _nx, _ny)
    if _np is None:
        _np = Generic.point.Point(_nx, _ny)
        layer.addObject(_np)
    _adjust_dimensions(layer, point, _np)
    layer.delObject(point)

def _move_segment(layer, segment, dx, dy):
    _p1, _p2 = segment.getEndpoints()
    _x, _y = _p1.getCoords()
    _nx = _x + dx
    _ny = _y + dy
    _np1 = layer.find('point', _nx, _ny)
    if _np1 is None:
        _np1 = Generic.point.Point(_nx, _ny)
        layer.addObject(_np1)
    _adjust_dimensions(layer, _p1, _np1)
    _x, _y = _p2.getCoords()
    _nx = _x + dx
    _ny = _y + dy
    _np2 = layer.find('point', _nx, _ny)
    if _np2 is None:
        _np2 = Generic.point.Point(_nx, _ny)
        layer.addObject(_np2)
    _adjust_dimensions(layer, _p2, _np2)
    segment.setP1(_np1)
    segment.setP2(_np2)
    layer.delObject(_p1)
    layer.delObject(_p2)

def _move_arc(layer, arc, dx, dy):
    _cp = arc.getCenter()
    _x, _y = _cp.getCoords()
    _ep1, _ep2 = arc.getEndpoints()
    _lp1 = layer.find('point', _ep1[0], _ep1[1])
    assert _lp1 is not None, "Lost arc first endpoint"
    _lp2 = layer.find('point', _ep2[0], _ep2[1])
    assert _lp2 is not None, "Lost arc second endpoint"
    _nx = _x + dx
    _ny = _y + dy
    _ncp = layer.find('point', _nx, _ny)
    if _ncp is None:
        _ncp = Generic.point.Point(_nx, _ny)
        layer.addObject(_ncp)
    for _ep in arc.getEndpoints():
        _ex, _ey = _ep
        _nx = _ex + dx
        _ny = _ey + dy
        _lp = layer.find('point', _nx, _ny)
        if _lp is None:
            _lp = Generic.point.Point(_nx, _ny)
            layer.addObject(_lp)
    arc.setCenter(_ncp)        
    _adjust_dimensions(layer, _cp, _ncp)
    layer.delObject(_cp)
    _lp1.freeUser(arc)
    layer.delObject(_lp1)
    _lp2.freeUser(arc)
    layer.delObject(_lp2)

def _move_circle(layer, circle, dx, dy): # moves circle and ccircles
    _cp = circle.getCenter()
    _x, _y = _cp.getCoords()
    _nx = _x + dx
    _ny = _y + dy
    _ncp = layer.find('point', _nx, _ny)
    if _ncp is None:
        _ncp = Generic.point.Point(_nx, _ny)
        layer.addObject(_ncp)
    _adjust_dimensions(layer, _cp, _ncp)
    circle.setCenter(_ncp)
    layer.delObject(_cp)

def _move_leader(layer, leader, dx, dy):
    _p1, _p2, _p3 = leader.getPoints()
    _x, _y = _p1.getCoords()
    _nx = _x + dx
    _ny = _y + dy
    _np1 = layer.find('point', _nx, _ny)
    if _np1 is None:
        _np1 = Generic.point.Point(_nx, _ny)
        layer.addObject(_np1)
    _adjust_dimensions(layer, _p1, _np1)
    _x, _y = _p2.getCoords()
    _nx = _x + dx
    _ny = _y + dy
    _np2 = layer.find('point', _nx, _ny)
    if _np2 is None:
        _np2 = Generic.point.Point(_nx, _ny)
        layer.addObject(_np2)
    _adjust_dimensions(layer, _p2, _np2)
    _x, _y = _p3.getCoords()
    _nx = _x + dx
    _ny = _y + dy
    _np3 = layer.find('point', _nx, _ny)
    if _np3 is None:
        _np3 = Generic.point.Point(_nx, _ny)
        layer.addObject(_np3)
    _adjust_dimensions(layer, _p3, _np3)
    leader.setP1(_np1)
    leader.setP2(_np2)
    leader.setP3(_np3)
    layer.delObject(_p1)
    layer.delObject(_p2)
    layer.delObject(_p3)

def _move_polyline(layer, polyline, dx, dy):
    _pts = polyline.getPoints()
    _npts = []
    for _pt in _pts:
        _x, _y = _pt.getCoords()
        _nx = _x + dx
        _ny = _y + dy
        _np = layer.find('point', _nx, _ny)
        if _np is None:
            _np = Generic.point.Point(_nx, _ny)
            layer.addObject(_np)
        _adjust_dimensions(layer, _pt, _np)
        _npts.append(_np)
    for _i in range(len(_npts)):
        polyline.setPoint(_i, _npts[_i])
    for _pt in _pts:
        layer.delObject(_pt)

def _move_segjoint(layer, segjoint, dx, dy):
    _s1, _s2 = segjoint.getSegments()
    _move_segment(layer, _s1, dx, dy)
    _move_segment(layer, _s2, dx, dy)

def _move_spcline(layer, spcline, dx, dy):
    _p = spcline.getLocation()
    _x, _y = _p.getCoords()
    _nx = _x + dx
    _ny = _y + dy
    _np = layer.find('point', _nx, _ny)
    if _np is None:
        _np = Generic.point.Point(_nx, _ny)
        layer.addObject(_np)
    _adjust_dimensions(layer, _p, _np)
    spcline.setLocation(_np)
    layer.delObject(_p)

def _move_cline(layer, cline, dx, dy):
    _p1, _p2 = cline.getKeypoints()
    _x, _y = _p1.getCoords()
    _nx = _x + dx
    _ny = _y + dy
    _np1 = layer.find('point', _nx, _ny)
    if _np1 is None:
        _np1 = Generic.point.Point(_nx, _ny)
        layer.addObject(_np1)
    _adjust_dimensions(layer, _p1, _np1)
    _x, _y = _p2.getCoords()
    _nx = _x + dx
    _ny = _y + dy
    _np2 = layer.find('point', _nx, _ny)
    if _np2 is None:
        _np2 = Generic.point.Point(_nx, _ny)
        layer.addObject(_np2)
    _adjust_dimensions(layer, _p2, _np2)
    cline.setP1(_np1)
    cline.setP2(_np2)
    layer.delObject(_p1)
    layer.delObject(_p2)

def _move_dimension(dim, dx, dy):
    _x, _y = dim.getLocation()
    _nx = _x + dx
    _ny = _y + dy
    dim.setLocation(_nx, _ny)

def move_objects(layer, objlist, dx, dy):
    """Move objects in a layer.

move_objects(layer, objlist, dx, dy)

layer; The layer in which all the objects are located
objlist: The list of objects to move. This argument must be a list
dx: The displacement along the x-axis
dy: The displacement along the y-axix
    """
    if not isinstance(layer, Generic.layer.Layer):
        raise TypeError, "Invalid layer: " + str(layer)
    if not isinstance(objlist, list):
        raise TypeError, "Invalid object list: " + str(objlist)
    _dx = dx
    if not isinstance(_dx, float):
        _dx = float(dx)
    _dy = dy
    if not isinstance(_dy, float):
        _dy = float(dy)
    if abs(_dx) < 1e-10 and abs(_dy) < 1e-10:
        return
    _objlist = []
    _objdict = {}
    for _obj in objlist:
        _lobj = layer.findObject(_obj)
        if _lobj is not _obj:
            raise ValueError, "Object not found in layer: " + str(_obj)
        _objdict[id(_obj)] = True
        _objlist.append(_obj)
    for _obj in _objlist:
        if isinstance(_obj, Generic.segment.Segment):
            _p1, _p2 = _obj.getEndpoints()
            _pid = id(_p1)
            if _pid in _objdict:
                _objdict[_pid] = False
            _pid = id(_p2)
            if _pid in _objdict:
                _objdict[_pid] = False
        elif isinstance(_obj, (Generic.circle.Circle,
                               Generic.arc.Arc,
                               Generic.ccircle.CCircle)):
            _cp = _obj.getCenter()
            _cpid = id(_cp)
            if _cpid in _objdict:
                _objdict[_cpid] = False
            if isinstance(_obj, Generic.arc.Arc):
                for _ep in _obj.getEndpoints():
                    _lp = layer.find('point', _ep[0], _ep[1])
                    if _lp is None:
                        raise ValueError, "Lost arc endpoint: " + str(_ep)
                    _pid = id(_lp)
                    if _pid in _objdict:
                        _objdict[_pid] = False
        elif isinstance(_obj, (Generic.hcline.HCLine,
                               Generic.vcline.VCLine,
                               Generic.acline.ACLine)):
            _p = _obj.getLocation()
            _pid = id(_p)
            if _pid in _objdict:
                _objdict[_pid] = False
        elif isinstance(_obj, Generic.cline.CLine):
            _p1, _p2 = _obj.getKeypoints()
            _pid = id(_p1)
            if _pid in _objdict:
                _objdict[_pid] = False
            _pid = id(_p2)
            if _pid in _objdict:
                _objdict[_pid] = False
        elif isinstance(_obj, (Generic.leader.Leader,
                               Generic.polyline.Polyline)):
            for _pt in _obj.getPoints():
                _pid = id(_pt)
                if _pid in _objdict:
                    _objdict[_pid] = False
        elif isinstance(_obj, (Generic.segjoint.Chamfer,
                               Generic.segjoint.Fillet)):
            _s1, _s2 = _obj.getSegments()
            _sid = id(_s1)
            if _sid in _objdict:
                _objdict[_sid] = False
            _p1, _p2 = _s1.getEndpoints()
            _pid = id(_p1)
            if _pid in _objdict:
                _objdict[_pid] = False
            _pid = id(_p2)
            if _pid in _objdict:
                _objdict[_pid] = False
            _sid = id(_s2)
            if _sid in _objdict:
                _objdict[_sid] = False
            _p1, _p2 = _s2.getEndpoints()
            _pid = id(_p1)
            if _pid in _objdict:
                _objdict[_pid] = False
            _pid = id(_p2)
            if _pid in _objdict:
                _objdict[_pid] = False
        else:
            pass
    for _obj in _objlist:
        if _objdict[id(_obj)]:
            if isinstance(_obj, Generic.point.Point):
                _move_point(layer, _obj, _dx, _dy)
            elif isinstance(_obj, Generic.segment.Segment):
                _move_segment(layer, _obj, _dx, _dy)
            elif isinstance(_obj, Generic.arc.Arc):
                _move_arc(layer, _obj, _dx, _dy)
            elif isinstance(_obj, (Generic.circle.Circle,
                                   Generic.ccircle.CCircle)):
                _move_circle(layer, _obj, _dx, _dy)
            elif isinstance(_obj, Generic.leader.Leader):
                _move_leader(layer, _obj, _dx, _dy)
            elif isinstance(_obj, Generic.polyline.Polyline):
                _move_polyline(layer, _obj, _dx, _dy)
            elif isinstance(_obj, (Generic.segjoint.Chamfer,
                                   Generic.segjoint.Fillet)):
                _move_segjoint(layer, _obj, _dx, _dy)
            elif isinstance(_obj, (Generic.hcline.HCLine,
                                   Generic.vcline.VCLine,
                                   Generic.acline.ACLine)):
                _move_spcline(layer, _obj, _dx, _dy)
            elif isinstance(_obj, Generic.cline.CLine):
                _move_cline(layer, _obj, _dx, _dy)
            elif isinstance(_obj, Generic.dimension.Dimension):
                _move_dimension(_obj, _dx, _dy)
            elif isinstance(_obj, Generic.text.TextBlock):
                _x, _y = _obj.getLocation()
                _nx = _x + _dx
                _ny = _y + _dy
                _obj.setLocation(_nx, _ny)
            else:
                pass



#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# a class for maintaining a sorted list
#
# each list holds only a single type of object, and
# the list is organized from smallest to largest
#

import baseobject

class TreeItem(baseobject.ModObject):
    """An object stored in a Tree.

A TreeItem object is meant to be a base class for classes
that will have instances stored in a Tree. TreeItem objects
are derived from ModObjects, so they share the same attributes
and methods as those objects.
    """
    def __init__(self):
        """Initialize a TreeItem object.

There are no arguments to this method.
        """
        baseobject.ModObject.__init__(self)
        self.__tree = None # private to the TreeItem

    def _setTree(self, tree):
        """Store a reference to the Tree this TreeItem is stored within.

_setTree(tree)

A TreeItem object can only be kept in one Tree at a time.

This routine is private to the TreeItem implementation.
        """
        if tree is None:
            self.__tree = None
        else:
            if not isinstance(tree, Tree):
                raise TypeError, "Invalid Tree for storage: " + str(tree)
            if self.__tree is not None:
                raise ValueError, "TreeItem already claimed by a tree."
            self.__tree = tree
        
    def modified(self):
        """Set the modified flag value to True.

modified()

This method extends the ModObject::modified() method
        """
        baseobject.ModObject.modified(self)
        if self.__tree is not None:
            self.__tree.markModified(self)

class Tree(list):
    """A class for Trees.

The Tree class is similar to a list, but it stores the
objects in a sorted order. A Tree object can be iterated
over, and there is a several method for scanning objects
held in the tree.

An object stored in the tree must be derived from the TreeItem
class.

A Tree has several methods:

clear(): Remove all objects from the tree.
store(obj): Store some object in the tree.
remove(obj): Remove an object from the tree
binscan(obj): Search the tree for the object using a binary search
scan(obj): Search the tree for the object using a linear search.
isModified(): An object stored in the Tree has its modified flag set to True
getModified(): Return the objects in the Tree that are modified
clean(): Sort the objects stored in the Tree, and remove any duplicates.

Trees also have a modified __add__ method that allows
you to add trees together, and a __contains__ method
for doing "obj in tree" or "obj not in tree" operations.

Also, a Tree supports indexed retrival similar to a list,
but not allow indexed assignment or indexed deletion.
    """
    def __init__(self, t):
        """Initialize a Tree.

Tree(t)

A Tree takes a single parameter t which is the
type of object stored within the tree. Any Tree
will only hold one type of object, and that object
type cannot be None.
        """
        if not issubclass(t, TreeItem):
            raise ValueError, "Invalid object for Tree storage: " + `t`
        list.__init__(self)
        self.__type = t
        self.__iter_index = 0
        self.__modified = []

    def __iter__(self):
        """Make a Tree iterable.
        """
        if len(self.__modified):
            raise ValueError, "Tree in modified state."
        self.__iter_index = 0
        return self

    def next(self):
        """This method is used for iteration.
        """
        _index = self.__iter_index
        _next = _index + 1
        if _next > len(self):
            raise StopIteration
        self.__iter_index = _next
        return self[_index]
    
    def __setitem__(self, index, value):
        raise StandardError, "Tree::__setitem__() is not supported."

    def __delitem__(self, index):
        raise StandardError, "Tree::__delitem__() is not supported."

    def __setslice__(self, i, j, vals):
        raise StandardError, "Tree::__setslice__() is not supported."

    def __delslice__(self, i, j):
        raise StandardError, "Tree::__delslice__() is not supported."

    def append(self, obj):
        raise StandardError, "Tree::append() is not supported."

    def insert(self, index, obj):
        raise StandardError, "Tree::insert() is not supported."

    def extend(self, objlist):
        raise StandardError, "Tree::extend() is not supported."

    def pop(self, idx=-1):
        raise StandardError, "Tree::pop() is not supported."

    def reverse(self):
        raise StandardError, "Tree::reverse() is not supported."

    def __contains__(self, obj):
        """Test if an object is in a Tree.

This method uses a binary search to find if an object
is in a Tree. It _should_ always be faster than a simple
linear search from first object to the last.
        """
        if not isinstance(obj, self.__type):
            raise TypeError, "Invalid object for inclusion test: " + `obj`
        if len(self.__modified):
            raise ValueError, "Tree in modified state."
        _lo = 0
        _hi = len(self)
        _scanlist = []
        _seen = False
        while _lo < _hi:
            _mid = (_hi+_lo)//2
            if _mid in _scanlist:
                break
            _scanlist.append(_mid)
            _res = cmp(self[_mid], obj)
            if _res == -1:
                _lo = _mid + 1
            elif _res == 1:
                _hi = _mid
            else:
                _seen = True
                break
        return _seen
        
    def clear(self):
        """Empty the Tree.

clear()
        """
        del self.__modified[:]
        list.__delslice__(self, 0, len(self))

    def markModified(self, treeitem):
        """Store a reference to a modified TreeItem object.

markModified(treeitem)

This routine is private to the Tree implementation.
        """
        self.__modified.append(treeitem)
        
    def store(self, obj):
        """Store an object in the Tree.

store(obj)

The object must be the same as was given when the tree
was instantiated. A TypeError is raised if another
object type is stored.

The objects in the Tree are searched by using the cmp()
function. Any class that will be stored in a Tree should
provide a suitable __cmp__ method.
        """
        if not isinstance(obj, self.__type):
            raise TypeError, "Invalid object for storage: " + `obj`
        obj._setTree(self)
        if len(self.__modified):
            raise ValueError, "Tree in modified state."
        _lo = 0
        _hi = len(self)
        _scanlist = []
        while _lo < _hi:
            _mid = (_hi+_lo)//2
            if _mid in _scanlist:
                break
            _scanlist.append(_mid)
            _res = cmp(self[_mid], obj)
            if _res == -1:
                _lo = _mid + 1
            elif _res == 1:
                _hi = _mid
            else:
                raise ValueError, "Equivalent object already in tree: " + `obj`
        list.insert(self, _lo, obj)
        assert _is_sorted(self), "Tree is not sorted: " + `self`

    def remove(self, obj):
        """Remove an object from a Tree.

remove(obj)

Delete the object from the Tree.
        """
        if not isinstance(obj, self.__type):
            raise TypeError, "Invalid object for removal: " + `obj`
        if len(self.__modified):
            raise ValueError, "Tree in modified state."
        obj._setTree(None)
        _lo = 0
        _hi = len(self)
        _scanlist = []
        while _lo < _hi:
            _mid = (_hi+_lo)//2
            if _mid in _scanlist:
                break
            _scanlist.append(_mid)
            _res = cmp(self[_mid], obj)
            if _res == -1:
                _lo = _mid + 1
            elif _res == 1:
                _hi = _mid
            else:
                list.__delitem__(self, _mid)
                break
    
    def binscan(self, obj, tol=0):
        """Scan the Tree for the an object using a binary search.

binscan(obj)

This method looks for an object in a Tree utilizing a simple
binary search. If the object is found within the tree, the
index of the object is returned. Otherwise, the function returns
None.
        """
        if not isinstance(obj, self.__type):
            raise TypeError, "Invalid object for binscan: " + `obj`
        if len(self.__modified):
            raise ValueError, "Tree in modified state."
        _lo = 0
        _hi = len(self)
        _scanlist = []
        _ro = None
        while _lo < _hi:
            _mid = (_hi+_lo)//2
            if _mid in _scanlist:
                break
            _scanlist.append(_mid)
            _res = cmp(self[_mid], obj)
            if _res == -1:
                _lo = _mid + 1
            elif _res == 1:
                _hi = _mid
            else:
                _ro = self[_mid]
                break
        return _ro

    def scan(self, obj, tol=0):
        """Scan the Tree for the an object from first to last.

scan(obj)

This method looks for an object in a Tree from the first object
to the last. It stops searching at the first instance which is
greater than the object, or the first instance where one object
is equal to the other.

The function returns None if no object in the Tree is found
equal to the object being compared against.
        """
        if not isinstance(obj, self.__type):
            raise TypeError, "Invalid object for scan: " + `obj`
        if len(self.__modified):
            raise ValueError, "Tree in modified state."
        _ro = None
        for _sobj in self:
            _res = cmp(_sobj, obj)
            if _res == 1:
                break
            if _res == 0:
                _ro = _sobj
                break
        return _ro

    def isModified(self):
        """Returns True if an object in the tree has the modified flag to True.

isModified()        
        """
        return len(self.__modified) != 0

    def getModified(self):
        """Return any objects in the tree that have the modified flag to True.

getModified()        
        """
        return self.__modified[:]
    
    def clean(self):
        """Sort the objects in the tree, and remove duplicated entities.

clean()

This function returns any duplicated entities found in the Tree.
It is up to the caller to deal with any returned objects.
        """
        _dups = []
        if len(self.__modified):
            self.sort()
            _pobj = self[0]
            _dlist = []
            for _i in range(1, len(self)):
                _obj = self[_i]
                if _obj == _pobj:
                    _dlist.append(_i)
                else:
                    _pobj = _obj
            if len(_dlist):
                _dlist.reverse()
                for _i in _dlist:
                    _dups.append(self[_i])
                    list.__delitem__(self, _i)
            for _obj in self.__modified:
                if _obj not in _dups:
                    _obj.reset()
            del self.__modified[:]
        return _dups

#
# this is a function used for assertion testing
#

def _is_sorted(tree):
    _res = True
    for _i in range(1, len(tree)):
        _tp = tree[_i-1]
        _ti = tree[_i]
        if _tp >= _ti:
            _res = False
            break
    return _res



#
# Copyright (c) 2002, 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# the event handling bits for construction lines
# and circles
#

import math

import pygtk
pygtk.require('2.0')
import gtk

import Generic.point
import Generic.segment
import Generic.circle
import Generic.arc
import Generic.hcline
import Generic.vcline
import Generic.acline
import Generic.cline
import Generic.ccircle
import Generic.util
import Generic.color

import gtkentities

#
# horizontal construction lines
#

def hcline_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _new_pt = gtkimage.findPoint(_x, _y, _tol)
    if _pt is not None:
        _x, _y = _pt.getCoords()
    tool.setPoint(_x, _y)
    gtkentities.create_entity(gtkimage, tool)

def hcline_entry_event_cb(gtkimage, widget, tool):
    _entry = gtkimage.getEntry()
    _text = _entry.get_chars(0,-1)
    _entry.delete_text(0,-1)
    if len(_text):
        _textval = eval(_text)
        _val = _textval
        if not isinstance(_val, float):
            _val = float(_textval)
        tool.setPoint(0.0, _val)
        gtkentities.create_entity(gtkimage, tool)
            
def hcline_mode_init(gtkimage, tool):
    gtkimage.setPrompt("Click in the drawing area or enter a point.")
    tool.setHandler("initialize", hcline_mode_init)
    tool.setHandler("button_press", hcline_button_press_cb)
    tool.setHandler("entry_event", hcline_entry_event_cb)

#
# vertical construction lines
#

def vcline_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _new_pt = gtkimage.findPoint(_x, _y, _tol)
    if _pt is not None:
        _x, _y = _pt.getCoords()
    tool.setPoint(_x, _y)
    gtkentities.create_entity(gtkimage, tool)

def vcline_entry_event_cb(gtkimage, widget, tool):
    _entry = gtkimage.getEntry()
    _text = _entry.get_chars(0,-1)
    _entry.delete_text(0,-1)
    if len(_text):
        _textval = eval(_text)
        _val = _textval
        if not isinstance(_val, float):
            _val = float(_textval)
        tool.setPoint(_val, 0.0)
        gtkentities.create_entity(gtkimage, tool)

def vcline_mode_init(gtkimage, tool):
    gtkimage.setPrompt("Click in the drawing area or enter x coordinate")
    tool.setHandler("initialize", vcline_mode_init)
    tool.setHandler("button_press", vcline_button_press_cb)
    tool.setHandler("entry_event", vcline_entry_event_cb)

#
# common functions for for acline, cline, and ccircle objects
#

def make_tuple(text):
    try:
        _tpl = eval(text)
    except:
        raise StandardError, "Invalid point: " + str(text)
    if not isinstance(_tpl, tuple):
        raise TypeError, "Invalid point: " + str(_tpl)
    if len(_tpl) != 2:
        raise ValueError, "Invalid point: " + str(_tpl)
    return _tpl
    
#
# angled construction lines
#

def acline_motion_notify_cb(gtkimage, widget, event, tool):
    _segs = []
    _ax, _ay = tool.getPoint()
    _pax, _pay = gtkimage.coordToPixTransform(_ax, _ay)
    _gc = gtkimage.getGC()
    _x = int(event.x)
    _y = int(event.y)
    _cp = tool.getCurrentPoint()
    if _cp is not None:
        _xc, _yc = _cp
        _segs.append((_pax, _pay, _xc, _yc))
    _xl, _yl = gtkimage.getPoint()
    tool.setLocation(_xl, _yl)
    tool.setCurrentPoint(_x, _y)
    _segs.append((_pax, _pay, _x, _y))
    widget.window.draw_segments(_gc, _segs)
        
def acline_entry_make_angle(gtkimage, widget, tool):
    _entry = gtkimage.getEntry()
    _text = _entry.get_chars(0,-1)
    _entry.delete_text(0,-1)
    if len(_text):
        _angle = Generic.util.make_angle(_text)
        tool.setAngle(_angle)
        gtkentities.create_entity(gtkimage, tool)

def acline_entry_make_pt(gtkimage, widget, tool):
    _entry = gtkimage.getEntry()
    _text = _entry.get_chars(0,-1)
    _entry.delete_text(0,-1)
    if len(_text):
        _x, _y = make_tuple(_text)
        tool.setPoint(_x, _y)
        tool.setHandler("button_press", acline_second_button_press_cb)
        tool.setHandler("entry_event", acline_entry_make_angle)
        tool.setHandler("motion_notify", acline_motion_notify_cb)
        gtkimage.setPrompt("Enter the angle or click in the drawing area")
        gtkimage.getGC().set_function(gtk.gdk.INVERT)

def acline_second_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _new_pt = gtkimage.findPoint(_x, _y, _tol)
    if _pt is not None:
        _x, _y = _pt.getCoords()
    tool.setLocation(_x, _y)
    gtkentities.create_entity(gtkimage, tool)

def acline_first_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _new_pt = gtkimage.findPoint(_x, _y, _tol)
    if _pt is not None:
        _x, _y = _pt.getCoords()
    tool.setPoint(_x, _y)
    tool.setHandler("button_press", acline_second_button_press_cb)
    tool.setHandler("entry_event", acline_entry_make_angle)
    tool.setHandler("motion_notify", acline_motion_notify_cb)
    gtkimage.setPrompt("Enter the angle or click in the drawing area")
    gtkimage.getGC().set_function(gtk.gdk.INVERT)

def acline_mode_init(gtkimage, tool):
    gtkimage.setPrompt("Click in the drawing area or enter a Point")
    tool.setHandler("initialize", acline_mode_init)
    tool.setHandler("button_press", acline_first_button_press_cb)
    tool.setHandler("entry_event", acline_entry_make_pt)

#
# two point construction line
#

def cline_motion_notify_cb(gtkimage, widget, event, tool):
    _segs = []
    _x1, _y1 = tool.getFirstPoint()
    _px1, _py1 = gtkimage.coordToPixTransform(_x1, _y1)
    _gc = gtkimage.getGC()
    _x = int(event.x)
    _y = int(event.y)
    _cp = tool.getCurrentPoint()
    if _cp is not None:
        _xc, _yc = _cp
        _segs.append((_px1, _py1, _xc, _yc))
    _ix, _iy = gtkimage.getPoint()
    tool.setSecondPoint(_ix, _iy)
    tool.setCurrentPoint(_x, _y)
    _segs.append((_px1, _py1, _x, _y))
    widget.window.draw_segments(_gc, _segs)

def cline_second_entry_make_pt(gtkimage, widget, tool):
    _entry = gtkimage.getEntry()
    _text = _entry.get_chars(0,-1)
    _entry.delete_text(0,-1)
    if len(_text):
        _x, _y = make_tuple(_text)
        tool.setSecondPoint(_x, _y)
        gtkentities.create_entity(gtkimage, tool)

def cline_first_entry_make_pt(gtkimage, widget, tool):
    _entry = gtkimage.getEntry()
    _text = _entry.get_chars(0,-1)
    _entry.delete_text(0,-1)
    if len(_text):
        _x, _y = make_tuple(_text)
        tool.setFirstPoint(_x, _y)
        tool.setHandler("button_press", cline_second_button_press_cb)
        tool.setHandler("motion_notify", cline_motion_notify_cb)
        tool.setHandler("entry_event", cline_second_entry_make_pt)
        gtkimage.setPrompt("Enter the second Point or click in the drawing area")
        gtkimage.getGC().set_function(gtk.gdk.INVERT)

def cline_second_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _new_pt = gtkimage.findPoint(_x, _y, _tol)
    if _pt is not None:
        _x, _y = _pt.getCoords()
    tool.setSecondPoint(_x, _y)
    gtkentities.create_entity(gtkimage, tool)

def cline_first_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _new_pt = gtkimage.findPoint(_x, _y, _tol)
    if _pt is not None:
        _x, _y = _pt.getCoords()
    tool.setFirstPoint(_x, _y)
    tool.setHandler("button_press", cline_second_button_press_cb)
    tool.setHandler("entry_event", cline_second_entry_make_pt)
    tool.setHandler("motion_notify", cline_motion_notify_cb)
    gtkimage.setPrompt("Enter the second point or click in the drawing area")
    gtkimage.getGC().set_function(gtk.gdk.INVERT)

def cline_mode_init(gtkimage, tool):
    gtkimage.setPrompt("Click in the drawing area or enter a Point")
    tool.setHandler("initialize", cline_mode_init)
    tool.setHandler("button_press", cline_first_button_press_cb)
    tool.setHandler("entry_event", cline_first_entry_make_pt)

#
# construction circles
#

def ccircle_make_point(gtkimage, text, addflag=True):
    _tpl = make_tuple(text)
    _x, _y = _tpl
    _pt = gtkimage.findPoint(_x, _y)
    if _pt is None:
        _pt = Generic.point.Point(_x, _y)
    return _pt

def ccircle_make_radius(text):
    _text_rad = eval(text)
    _rad = _text_rad
    if not isinstance(_rad, float):
        _rad = float(_text_rad)
    if _rad < 1e-10:
        raise ValueError, "Invalid radius: %g" % _rad
    return _rad

def ccircle_motion_notify_cb(gtkimage, widget, event, tool):
    _tx, _ty = tool.getLocation()
    _px, _py = gtkimage.coordToPixTransform(_tx, _ty)
    _width, _height = gtkimage.getSize()
    _gc = gtkimage.getGC()
    _x = int(event.x)
    _y = int(event.y)
    _cp = tool.getCurrentPoint()
    if _cp is not None:
        _xc, _yc = _cp
        _radius = math.hypot((_px - _xc),(_py - _yc))
        _xmin = _px - _radius
        _ymin = _py - _radius
        _cw = _ch = _radius * 2
        widget.window.draw_arc(_gc, gtk.FALSE, _xmin, _ymin, _cw, _ch,
                               0, 360*64)
    tool.setCurrentPoint(_x, _y)
    _radius = math.hypot((_px - _x),(_py - _y))
    _xmin = _px - _radius
    _ymin = _py - _radius
    _cw = _ch = _radius * 2
    widget.window.draw_arc(_gc, gtk.FALSE, _xmin, _ymin, _cw, _ch, 0, 360*64)

def ccircle_entry_make_radius(gtkimage, widget, tool):
    entry = gtkimage.getEntry()
    text = entry.get_chars(0,-1)
    entry.delete_text(0,-1)
    if len(text):
        try:
            radius =  ccircle_make_radius(text)
        except:
            print "Invalid radius: " + `text` # need better error handling
            return
        tool.pushObject(radius)
        ccircle_create(gtkimage, tool)

def ccircle_create(gtkimage, tool):
    assert len(tool) == 2, "CCircle tool should hold two objects."
    _obj = tool.popObject() # radius or a Point
    _cen = tool.popObject() # center Point
    if isinstance(_obj, float):
        _radius = _obj
    elif isinstance(_obj, Generic.point._Point):
        _radius = _obj - _cen
    else:
        raise "Unexpected object type: " + `_obj`
    _active_layer = gtkimage.getActiveLayer()
    gtkimage.startAction()
    try:
        if _active_layer.findObject(_cen) is None:
            _active_layer.addObject(_cen)
        gtkimage.addObject(Generic.ccircle.CCircle(_cen, _radius))
    finally:
        gtkimage.endAction()
    gtkimage.redraw()
    gtkimage.setPrompt("Click in drawing area or enter a Point")
    init_func = tool.getHandler("initialize")
    init_func(tool)

def ccircle_entry_make_pt(gtkimage, widget, tool):
    _entry = gtkimage.getEntry()
    _text = _entry.get_chars(0,-1)
    _entry.delete_text(0,-1)
    if len(_text):
        try:
            _p = ccircle_make_point(gtkimage, _text)
        except:
            print "Invalid point: " + `_text` # need better error handling
            return
        tool.pushObject(_p)
        if len(tool) == 1:
            _x, _y = _p.getCoords()
            tool.storeCoords(_x, _y)
            tool.setLocation(_x, _y)
            tool.setHandler("motion_notify", ccircle_motion_notify_cb)
            tool.setHandler("entry_event", ccircle_entry_make_radius)
            gtkimage.setPrompt("Enter the radius or click in the drawing area")
            gtkimage.getGC().set_function(gtk.gdk.INVERT)
        else:
            ccircle_create(gtkimage, tool)

def ccircle_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _new_pt = gtkimage.findPoint(_x, _y, _tol)
    if _pt is None:
        _pt = Generic.point.Point(_x, _y)
    tool.pushObject(_pt)
    if len(tool) == 1:
        _x, _y = _pt.getCoords()
        tool.storeCoords(_x, _y)
        tool.setLocation(_x, _y)
        tool.setHandler("motion_notify", ccircle_motion_notify_cb)
        tool.setHandler("entry_event", ccircle_entry_make_radius)
        gtkimage.setPrompt("Enter the radius or click in the drawing area")
        gtkimage.getGC().set_function(gtk.gdk.INVERT)
    else:
        ccircle_create(gtkimage, tool)

def ccircle_cpmode_init(gtkimage, tool):
    gtkimage.setPrompt("Click in the drawing area or enter a Point")
    tool.setHandler("button_press", gtkentities.circle_center_button_press_cb)
    tool.setHandler("entry_event", gtkentities.circle_point_entry_event_cb)
    tool.setHandler("initialize", ccircle_cpmode_init)

#
# two-point construction circle
#

def ccircle_tp_create(gtkimage, tool):
    assert len(tool) == 2, "CCircle TP tool should hold two objects."
    _p2 = tool.popObject()
    _p1 = tool.popObject()
    _xc = (_p1.x + _p2.x)/2.0
    _yc = (_p1.y + _p2.y)/2.0
    _rad = (_p2 - _p1)/2.0
    _pt, _new_pt = gtkimage.findPoint(_xc, _yc)
    gtkimage.startAction()
    try:
        if _pt is None:
            _pt = Generic.point.Point(_xc, _yc)
            gtkimage.addObject(_pt)
        gtkimage.addObject(Generic.ccircle.CCircle(_pt, _rad))
    finally:
        gtkimage.endAction()
    gtkimage.redraw()
    gtkimage.setPrompt("Click in drawing area or enter a Point")
    init_func = tool.getHandler("initialize")
    init_func(tool)


def ccircle_tp_motion_notify_cb(gtkimage, widget, event, tool):
    _tx, _ty = tool.getCoordPair(0)
    _px, _py = gtkimage.coordToPixTransform(_tx, _ty)
    _gc = gtkimage.getGC()
    _x = int(event.x)
    _y = int(event.y)
    _cp = tool.getCurrentPoint()
    if _cp is not None:
        _xc, _yc = _cp
        _xcen = _px + (_xc - _px)//2
        _ycen = _py + (_yc - _py)//2
        _radius = math.hypot((_xc - _xcen), (_yc - _ycen))
        _xmin = _xcen - _radius
        _ymin = _ycen - _radius
        _cw = _ch = _radius * 2
        widget.window.draw_arc(_gc, gtk.FALSE, _xmin, _ymin, _cw, _ch,
                               0, 360*64)
    tool.setCurrentPoint(_x, _y)
    _xcen = _px + (_x - _px)//2
    _ycen = _py + (_y - _py)//2
    _radius = math.hypot((_x - _xcen), (_y - _ycen))
    _xmin = _xcen - _radius
    _ymin = _ycen - _radius
    _cw = _ch = _radius * 2
    widget.window.draw_arc(_gc, gtk.FALSE, _xmin, _ymin, _cw, _ch, 0, 360*64)

def ccircle_tp_entry_make_pt(gtkimage, widget, tool):
    _entry = gtkimage.getEntry()
    _text = _entry.get_chars(0,-1)
    _entry.delete_text(0,-1)
    if len(_text):
        try:
            _p = ccircle_make_point(gtkimage, _text, False)
        except:
            print "Invalid point: " + `_text` # need better error handling
            return
        tool.pushObject(_p)
        if len(tool) == 1:
            _x, _y = _p.getCoords()
            tool.storeCoords(_x, _y)
            tool.setLocation(_x, _y)
            tool.setHandler("motion_notify", ccircle_tp_motion_notify_cb)
            gtkimage.setPrompt("Enter the second point or click in the drawing area")
            gtkimage.getGC().set_function(gtk.gdk.INVERT)
        else:
            ccircle_tp_create(gtkimage, tool)

def ccircle_tp_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _new_pt = gtkimage.findPoint(_x, _y, _tol)
    if _pt is None:
        _pt = Generic.point.Point(_x, _y)
    tool.pushObject(_pt)
    if len(tool) == 1:
        _x, _y = _pt.getCoords()
        tool.storeCoords(_x, _y)
        tool.setLocation(_x, _y)
        tool.setHandler("motion_notify", ccircle_tp_motion_notify_cb)
        gtkimage.setPrompt("Enter the second point or click in the drawing area")
        gtkimage.getGC().set_function(gtk.gdk.INVERT)
    else:
        ccircle_tp_create(gtkimage, tool)

def ccircle_tpmode_init(gtkimage, tool):
    gtkimage.setPrompt("Click in the drawing area or enter a Point")
    tool.setHandler("button_press", gtkentities.circle_tp_first_button_press_cb)
    tool.setHandler("entry_event", gtkentities.circle_tp_first_entry_event_cb)
    tool.setHandler("initialize", ccircle_tpmode_init)

#
# perpendicular construction line creation
#

def perp_cline_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _active_layer = gtkimage.getActiveLayer()
    _hits = _active_layer.mapPoint((_x, _y), _tol, 1)
    if len(_hits):
        _obj, _pt = _hits[0]
        _lp = _active_layer.findObject(_pt)
        _add_pt = False
        _pcl = None
        if _lp is None:
            _lp = _pt
            _add_pt = True
        if isinstance(_obj, Generic.segment.Segment):
            _p1, _p2 = _obj.getEndpoints()
            _p1x, _p1y = _p1.getCoords()
            _p2x, _p2y = _p2.getCoords()
            if abs(_p1x - _p2x) < 1e-10: # vertical
                _pcl = Generic.hcline.HCLine(_lp)
            elif abs(_p1y - _p2y) < 1e-10: # horizontal
                _pcl = Generic.vcline.VCLine(_lp)
            else:
                _slope = (180.0/math.pi) * math.atan2((_p2y - _p1y),
                                                      (_p2x - _p1x)) + 90.0
                _pcl = Generic.acline.ACLine(_lp, _slope)
        elif isinstance(_obj, (Generic.circle.Circle,
                               Generic.arc.Arc,
                               Generic.ccircle.CCircle)):
            _cp = _obj.getCenter()
            _pcl = Generic.cline.CLine(_cp, _lp)
        elif isinstance(_obj, Generic.hcline.HCLine):
            _pcl = Generic.vcline.VCLine(_lp)
        elif isinstance(_obj, Generic.vcline.VCLine):
            _pcl = Generic.hcline.HCLine(_lp)
        elif isinstance(_obj, Generic.acline.ACLine):
            _angle = _obj.getAngle()
            if abs(_angle) < 1e-10: # horizontal
                _pcl = Generic.vcline.VCLine(_lp)
            elif abs(abs(_angle) - 90.0) < 1e-10: # vertical
                _pcl = Generic.hcline.HCLine(_lp)
            else:
                _slope = _angle + 90.0
                _pcl = Generic.acline.ACLine(_lp, _slope)
        elif isinstance(_obj, Generic.cline.CLine):
            _p1, _p2 = _obj.getKeypoints()
            _p1x, _p1y = _p1.getCoords()
            _p2x, _p2y = _p2.getCoords()
            if abs(_p1x - _p2x) < 1e-10: # vertical
                _pcl = Generic.hcline.HCLine(_lp)
            elif abs(_p1y - _p2y) < 1e-10: # horizontal
                _pcl = Generic.vcline.VCLine(_lp)
            else:
                _slope = (180.0/math.pi) * math.atan2((_p2y - _p1y),
                                                      (_p2x - _p1x)) + 90.0
                _pcl = Generic.acline.ACLine(_lp, _slope)
        else:
            pass
        gtkimage.startAction()
        try:
            if _add_pt:
                _active_layer.addObject(_lp)
            if _pcl is not None:
                _active_layer.addObject(_pcl)
        finally:
            gtkimage.endAction()
        if _add_pt or _pcl is not None:
            gtkimage.redraw()

def perpendicular_cline_mode_init(tool):
    tool.initialize()
    tool.setHandler("button_press", perp_cline_button_press_cb)

#
# tangent cline creation
#

def tangent_cline_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _new_pt = gtkimage.findPoint(_x, _y, _tol)
    if _pt is not None:
        _x, _y = _pt.getCoords()
        _layer = gtkimage.getActiveLayer()
        _rtd = 180.0/math.pi
        _cobj = None
        _angle = None
        _circles = (_layer.getLayerEntities("circle") +
                    _layer.getLayerEntities("ccircle"))
        for _circle in _circles:
            _cx, _cy = _circle.getCenter().getCoords()
            _rad = _circle.getRadius()
            if abs(math.hypot((_x - _cx), (_y - _cy)) - _rad) < 1e-10:
                _cobj = _circle
                _angle = _rtd * math.atan2((_y - _cy), (_x - _cx))
                if _angle < 0.0:
                    _angle = _angle + 360.0
                break
        if _cobj is None:
            for _arc in _layer.getLayerEntities("arc"):
                _cx, _cy = _arc.getCenter().getCoords()
                _rad = _arc.getRadius()
                if abs(math.hypot((_cx - _x), (_cy - _y)) - _rad) < 1e-10:
                    _angle = _rtd * math.atan2((_y - _cy), (_x - _cx))
                    if _angle < 0.0:
                        _angle = _angle + 360.0
                    if _arc.throughAngle(_angle):
                        _cobj = _arc
                        break
        if _cobj is not None:
            gtkimage.startAction()
            try:
                if _new_pt:
                    _layer.addObject(_pt)
                if (abs(_angle) < 1e-6 or
                    abs(_angle - 180.0) < 1e-6 or
                    abs(_angle - 360.0) < 1e-6):
                    _tcl = Generic.vcline.VCLine(_pt)
                elif (abs(_angle - 90.0) < 1e-6 or
                      abs(_angle - 270.0) < 1e-6):
                    _tcl = Generic.hcline.HCLine(_pt)
                else:
                    _slope = _angle + 90.0
                    _tcl = Generic.acline.ACLine(_pt, _slope)
                _layer.addObject(_tcl)
            finally:
                gtkimage.endAction()
            gtkimage.redraw()

def tangent_cline_mode_init(tool):
    tool.initialize()
    tool.setHandler("button_press", tangent_cline_button_press_cb)

#
# parallel offset mode
#

def parallel_refpt_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    tool.setReferencePoint(_x, _y)
    _init_func = tool.getHandler("initialize")
    tool.create(gtkimage)
    gtkimage.redraw()
    _init_func(gtkimage, tool)

def parallel_conline_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _objdict = gtkimage.mapPoint(_x, _y, _tol, 1)
    if len(_objdict):
        _active_layer = gtkimage.getActiveLayer()
        if _active_layer in _objdict:
            for _obj, _pt in _objdict[_active_layer]:
                if isinstance(_obj, (Generic.hcline.HCLine,
                                     Generic.vcline.VCLine,
                                     Generic.acline.ACLine,
                                     Generic.cline.CLine)):
                    tool.setConstructionLine(_obj)
                    tool.setHandler("button_press", parallel_refpt_button_press_cb)
                    gtkimage.setPrompt("Click on the side to place the new construction line.")

def parallel_second_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _flag = gtkimage.findPoint(_x, _y, _tol)
    if _pt is not None:
        _x, _y = _pt.getCoords()
    _x1, _y1 = tool.getLocation()
    _offset = math.hypot((_x - _x1), (_y - _y1))
    tool.setOffset(_offset)
    tool.setHandler("button_press", parallel_conline_button_press_cb)
    gtkimage.setPrompt("Click on the reference construction line.")

def parallel_first_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _flag = gtkimage.findPoint(_x, _y, _tol)
    if _pt is not None:
        _x, _y = _pt.getCoords()
    tool.setLocation(_x, _y)
    tool.setHandler("button_press", parallel_second_button_press_cb)
    tool.delHandler("entry_event")
    gtkimage.setPrompt("Click another point to define the offset distance.")

def parallel_entry_event(gtkimage, widget, tool):
    _entry = gtkimage.getEntry()
    _text = _entry.get_chars(0,-1)
    _entry.delete_text(0,-1)
    if len(_text):
        _textdist = eval(_text)
        _dist = _textdist
        if not isinstance(_dist, float):
            _dist = float(_textdist)
        tool.setOffset(_dist)
        tool.delHandler("entry_event")
        tool.setHandler("button_press", parallel_conline_button_press_cb)
        gtkimage.setPrompt("Click on the reference construction line.")

def parallel_offset_mode_init(gtkimage, tool):
    gtkimage.setPrompt("Enter the distance or click in the drawing area.")
    tool.setHandler("initialize", parallel_offset_mode_init)
    tool.setHandler("button_press", parallel_first_button_press_cb)
    tool.setHandler("entry_event", parallel_entry_event)

#
# construction circle tangent to a construction line
#

def ccircle_single_second_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _flag = gtkimage.findPoint(_x, _y, _tol)
    if _pt is not None:
        _x, _y = _pt.getCoords()
    tool.setLocation(_x, _y)
    gtkentities.create_entity(gtkimage, tool)

def ccircle_single_motion_notify_cb(gtkimage, widget, event, tool):
    _gc = gtkimage.getGC()
    _upp = gtkimage.getUnitsPerPixel()
    _rect = tool.getPixelRect()
    if _rect is not None:
        _xmin, _ymin, _width, _height = _rect
        widget.window.draw_arc(_gc, gtk.FALSE, _xmin, _ymin, _width, _height,
                               0, 360*64)
    _ix, _iy = gtkimage.getPoint()
    tool.setLocation(_ix, _iy)
    _cx, _cy = tool.getCenter()
    _radius = tool.getRadius()
    _pcx, _pcy = gtkimage.coordToPixTransform(_cx, _cy)
    _pr = int(_radius/_upp)
    _xmin = _pcx - _pr
    _ymin = _pcy - _pr
    _width = _height = _pr * 2
    tool.setPixelRect(_xmin, _ymin, _width, _height)
    widget.window.draw_arc(_gc, gtk.FALSE, _xmin, _ymin, _width, _height,
                           0, 360*64)

def ccircle_single_first_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _objdict = gtkimage.mapPoint(_x, _y, _tol, 1)
    if len(_objdict):
        _active_layer = gtkimage.getActiveLayer()
        if _active_layer in _objdict:
            for _obj, _pt in _objdict[_active_layer]:
                if isinstance(_obj, (Generic.hcline.HCLine,
                                     Generic.vcline.VCLine,
                                     Generic.acline.ACLine,
                                     Generic.cline.CLine,
                                     Generic.ccircle.CCircle)):
                    tool.setConstructionLine(_obj)
                    tool.setHandler("button_press", ccircle_single_second_button_press_cb)
                    tool.setHandler("motion_notify", ccircle_single_motion_notify_cb)
                    gtkimage.setPrompt("Click where the circle should be drawn.")
                    gtkimage.getGC().set_function(gtk.gdk.INVERT)

def tangent_ccircle_mode_init(gtkimage, tool):
    gtkimage.setPrompt("Click on the construction object used for tangency.")
    tool.setHandler("initialize", tangent_ccircle_mode_init)
    tool.setHandler("button_press", ccircle_single_first_button_press_cb)

#
# construction circle between two construction lines
#

def two_cline_set_circle_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _flag = gtkimage.findPoint(_x, _y, _tol)
    if _pt is not None:
        _x, _y = _pt.getCoords()
    tool.setLocation(_x, _y)
    gtkentities.create_entity(gtkimage, tool)

def two_cline_motion_notify_cb(gtkimage, widget, event, tool):
    _gc = gtkimage.getGC()
    _upp = gtkimage.getUnitsPerPixel()
    _rect = tool.getPixelRect()
    if _rect is not None:
        _xmin, _ymin, _width, _height = _rect
        widget.window.draw_arc(_gc, gtk.FALSE, _xmin, _ymin, _width, _height,
                               0, 360*64)
    _ix, _iy = gtkimage.getPoint()
    tool.setLocation(_ix, _iy)
    _radius = tool.getRadius()
    if _radius > 0.0:
        _cx, _cy = tool.getCenter()
        _pcx, _pcy = gtkimage.coordToPixTransform(_cx, _cy)
        _pr = int(_radius/_upp)
        _xmin = _pcx - _pr
        _ymin = _pcy - _pr
        _width = _height = _pr * 2
        tool.setPixelRect(_xmin, _ymin, _width, _height)
        widget.window.draw_arc(_gc, gtk.FALSE, _xmin, _ymin, _width, _height,
                               0, 360*64)
    
def two_cline_second_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _objdict = gtkimage.mapPoint(_x, _y, _tol, 1)
    if len(_objdict):
        _active_layer = gtkimage.getActiveLayer()
        if _active_layer in _objdict:
            _first_conobj = tool.getFirstConObject()
            for _obj, _pt in _objdict[_active_layer]:
                if _obj is _first_conobj:
                    continue
                if isinstance(_obj, (Generic.hcline.HCLine,
                                     Generic.vcline.VCLine,
                                     Generic.acline.ACLine,
                                     Generic.cline.CLine)):
                    tool.setHandler("button_press", two_cline_set_circle_cb)
                    tool.setHandler("motion_notify", two_cline_motion_notify_cb)
                    tool.setSecondConObject(_obj)
                    gtkimage.setPrompt("Click where you want the tangent circle to be.")
                    gtkimage.getGC().set_function(gtk.gdk.INVERT)

def two_cline_first_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _objdict = gtkimage.mapPoint(_x, _y, _tol, 1)
    if len(_objdict):
        _active_layer = gtkimage.getActiveLayer()
        if _active_layer in _objdict:
            for _obj, _pt in _objdict[_active_layer]:
                if isinstance(_obj, (Generic.hcline.HCLine,
                                     Generic.vcline.VCLine,
                                     Generic.acline.ACLine,
                                     Generic.cline.CLine,
                                     Generic.ccircle.CCircle)):
                    tool.setHandler("button_press", two_cline_second_button_press_cb)
                    tool.setFirstConObject(_obj)
                    gtkimage.setPrompt("Click on the second construction line for tangency.")

def two_cline_tancc_mode_init(gtkimage, tool):
    gtkimage.setPrompt("Click on the first construction line for tangency.")
    tool.setHandler("initialize", two_cline_tancc_mode_init)
    tool.setHandler("button_press", two_cline_first_button_press_cb)

#
# tangent lines around two circles
#

def two_ccircle_tangent_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    tool.setLocation(_x, _y)
    gtkentities.create_entity(gtkimage, tool)    

def _draw_two_circle_tangents(gtkimage, tool):
    _tanpts = tool.getTangentPoints()
    assert len(_tanpts), "No tangent points defined!"
    _gc = gtkimage.getGC()
    _da = gtkimage.getDA()
    #
    # adjust the GC to draw the temporary segments in
    # a distinctive manner
    #
    _gc.set_line_attributes(1, gtk.gdk.LINE_DOUBLE_DASH,
                            gtk.gdk.CAP_BUTT, gtk.gdk.JOIN_MITER)
    _gc.set_dashes(0, [3, 3])
    _cstring = '#ffff99' # yellowish color
    _tempcolor = Generic.color.Color(_cstring)
    if gtkimage.hasGTKColor(_tempcolor):
        _color = gtkimage.getGTKColor(_tempcolor)
    else:
        _color = _da.get_colormap().alloc_color(_cstring)
        gtkimage.saveGTKColor(_tempcolor, _color)
    _gc.set_foreground(_color)
    _gc.set_function(gtk.gdk.COPY)
    #
    _segs = []
    for _set in _tanpts:
        _x1, _y1, _x2, _y2 = _set
        # print "x1: %g; y1: %g" % (_x1, _y1)
        # print "x2: %g; y2: %g" % (_x2, _y2)
        _px1, _py1 = gtkimage.coordToPixTransform(_x1, _y1)
        _px2, _py2 = gtkimage.coordToPixTransform(_x2, _y2)
        _segs.append((_px1, _py1, _px2, _py2))
    _da.window.draw_segments(_gc, _segs)

def two_circle_tangent_second_button_press_cb(gtkimage, widget, event, tool):
    # print "in second_button_press_cb() ..."
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _objdict = gtkimage.mapPoint(_x, _y, _tol, 1)
    if len(_objdict):
        _active_layer = gtkimage.getActiveLayer()
        if _active_layer in _objdict:
            for _obj, _pt in _objdict[_active_layer]:
                if isinstance(_obj, Generic.ccircle.CCircle):
                    tool.setHandler("button_press", two_circle_tangent_second_button_press_cb)
                    tool.setSecondCCircle(_obj)
                    if tool.hasTangentPoints():
                        _draw_two_circle_tangents(gtkimage, tool)
                        gtkimage.setPrompt("Click on the segment to keep.")
                        tool.setHandler("button_press", two_ccircle_tangent_cb)
                    else:
                        tool.reset()
                        two_circle_tangent_line_mode_init(gtkimage, tool)

def two_circle_tangent_first_button_press_cb(gtkimage, widget, event, tool):
    # print "in first_button_press_cb() ..."
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _objdict = gtkimage.mapPoint(_x, _y, _tol, 1)
    if len(_objdict):
        _active_layer = gtkimage.getActiveLayer()
        if _active_layer in _objdict:
            for _obj, _pt in _objdict[_active_layer]:
                if isinstance(_obj, Generic.ccircle.CCircle):
                    tool.setHandler("button_press", two_circle_tangent_second_button_press_cb)
                    tool.setFirstCCircle(_obj)
                    gtkimage.setPrompt("Click on the second construction circle.")

def two_circle_tangent_line_mode_init(gtkimage, tool):
    gtkimage.setPrompt("Click on the first construction circle.")
    tool.setHandler("initialize", two_circle_tangent_line_mode_init)
    tool.setHandler("button_press", two_circle_tangent_first_button_press_cb)



#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# linetype class
#

import types
from sys import maxint

import globals

class Linetype(object):
    """A class representing linetypes.

This class provides the basis for solid and dashed lines.
Dashed lines are more "interesting" in the various ways
the dashes can be drawn.

A Linetype object has two attributes:

name: A string describing this linetype
dlist: A list of integers used for drawing dashed lines

A solid line has a dlist attribute of None.

A Linetype object has the following methods:

getName(): Return the Linetype name
getList(): Return the Linetype dashlist
clone(): Make a copy of a Linetype object.
    """
    def __init__(self, name, dashlist=None):
        """Initstantiate a Linetype object.

Linetype(name, dashlist)

The name should be a string representing what this linetype
is called. The dashlist can be None (solid lines) or a list of
integers which will be used to determine the on/off bits
of a dashed line. The list must have at least two integers.
        """
        _n = name
        if not isinstance(_n, types.StringTypes):
            raise TypeError, "Invalid Linetype name: " + `_n`
        if not isinstance(_n, unicode):
            _n = unicode(name)
        _l = dashlist
        if _l is not None:
            if not isinstance(_l, list):
                raise TypeError, "Second argument must be a list or None."
            _length = len(_l)
            if _length < 2:
                raise ValueError, "Dash list must hold at least two integers."
            _temp = []
            for _i in range(_length):
                _item = _l[_i]
                if not isinstance(_item, int):
                    _item = int(_l[_i])
                if _item < 0:
                    raise ValueError, "Invalid list value: %d" % _item
                _temp.append(_item)
            _l = _temp
        self.__name = _n
        if _l is None:
            self.__list = None
        else:
            self.__list = _l[:]

    def __str__(self):
        if self.__list is None:
            return "Solid Line: '%s'" % self.__name
        else:
            str = "Dashed Line: '%s': " % self.__name
            return str + `self.__list`
        
    def __eq__(self, obj):
        """Compare one Linetype to another for equality.

The comparison examines the dash list - if both lists
are None, or the same length and with identical values,
the function returns True. Otherwise the function returns
False.
        """
        if not isinstance(obj, Linetype):
            raise TypeError, "Invalid Linetype comparison: " + `obj`
        _val = False
        _objlist = obj.getList()
        if self.__list is None and _objlist is None:
            _val = True
        elif self.__list is not None and _objlist is not None:
            if self.__list == _objlist:
                _val = True
        return _val

    def __ne__(self, obj):
        """Compare one Linetype to another for non-equality.

The comparison examines the dash list - if both lists
are None, or the same length and with identical values,
the function returns False. Otherwise the function returns
True.
        """
        if not isinstance(obj, Linetype):
            raise TypeError, "Invalid Linetype comparison: " + `obj`
        _val = True
        _objlist = obj.getList()
        if self.__list is None and _objlist is None:
            _val = False
        elif self.__list is not None and _objlist is not None:
            if self.__list == _objlist:
                _val = False
        return _val

    def __hash__(self):
        """Provide a hash function for Linetypes.

Defining this method allows Linetype objects to be stored
as dictionary keys.
        """
        _dashlist = self.__list
        _val = 0 # hash value for solid lines
        if _dashlist is not None:
            _val = hash_dashlist(_dashlist)
        return _val
    
    def getName(self):
        """Get the name of the Linetype.

getName()        
        """
        return self.__name

    name = property(getName, None, None, "Linetype name.")

    def getList(self):
        """Get the list used for determining the dashed line pattern.

getList()
This function returns None for solid Linetypes.
        """
        _list = None
        if self.__list is not None:
            _list = self.__list[:]
        return _list

    list = property(getList, None, None, "Linetype dash list.")

    def clone(self):
        """Make a copy of a Linetype

clone()        
        """
        _name = self.__name[:]
        _dashlist = None
        if self.__list is not None:
            _dashlist = self.__list[:]
        return Linetype(_name, _dashlist)

#
# LinetypeDict Class
#
# The LinetypeDict is built from the dict object. Using instances
# of this class will guarantee than only Linetype objects will be
# stored in the instance
#

class LinetypeDict(dict):
    def __init__(self):
        dict.__init__(self)
        
    def __setitem__(self, key, value):
        if not isinstance(key, Linetype):
            raise TypeError, "LinetypeDict keys must be Linetype objects: " + `key`
        if not isinstance(value, Linetype):
            raise TypeError, "LinetypeDict values must be Linetype objects: " + `value`
        dict.__setitem__(self, key, value)

#
# the hash function for linetype dashlists
#

def hash_dashlist(dashlist):
    """The hashing function for linetype dashlists

hash_dashlist(dashlist)

Argument "dashlist" should be a list containing two or
more integer values.
    """
    if not isinstance(dashlist, list):
        raise TypeError, "Invalid list: " + `dashlist`
    if len(dashlist) < 2:
        raise ValueError, "Invalid list length: " + str(dashlist)
    _dlist = []
    for _obj in dashlist:
        if not isinstance(_obj, int):
            raise TypeError, "Invalid list item: " + `_obj`
        if _obj < 0:
            raise ValueError, "Invalid list value: %d" % _obj
        _dlist.append(_obj)
    _val = (0xffffff & _dlist[0]) << 6
    for _i in _dlist:
        _val =  c_mul(160073, _val) ^ ((_i << 5) | _i) # made this up
    _val = _val ^ len(_dlist)
    if _val == -1:
        _val = -2
    return _val

#
# mulitplication used for hashing
#
# an eval-based routine came from http://effbot.org/zone/python-hash.htm,
# but the move to Python 2.3 printed warnings when executing that code
# due to changes in large hex values, so the code was re-written to
# avoid the eval-loop and hackish long->str->int conversions ...
#

def c_mul(a, b):
    _lval = (long(a * b) & 0xffffffffL) - maxint
    return int(_lval)

#
# find an existing linetype in the global linetype dictionary
#

def get_linetype_by_dashes(dashlist):
    if dashlist is None:
        _key = 0
    elif isinstance(dashlist, list):
        _key = hash_dashlist(dashlist)
    else:
        raise TypeError, "Invalid dashlist: " + `dashlist`
    return globals.linetypes.get(_key)
        
def get_linetype_by_name(name):
    if not isinstance(name, types.StringTypes):
        raise TypeError, "Invalid linetype name: " + str(name)
    _name = name
    if not isinstance(_name, unicode):
        _name = unicode(name)
    for _lt in globals.linetypes:
        if _lt.getName() == _name:
            return _lt
    return None



from bundlebuilder import buildapp
from plistlib import Plist, Dict


plist = Plist(
    CFBundleDocumentTypes = [
        Dict(
            CFBundleTypeExtensions = ["xml", "xml.gz", "*"],
            CFBundleTypeName = "XML File",
            CFBundleTypeRole = "Editor",
            NSDocumentClass = "ImageDocument",
        ),
        Dict(
            CFBundleTypeExtensions = ["dwg"],
            CFBundleTypeName = "DWG File",
            CFBundleTypeRole = "Viewer",
            NSDocumentClass = "ImageDocument",
        ),
    ]
)


buildapp(
	mainprogram = "PythonCad.py",
	resources = ["Interface/Cocoa/MainMenu.nib", "Interface/Cocoa/ImageDocument.nib", "Generic", "Interface", "prefs.py"],
	nibname = "MainMenu",
    plist = plist,
)
